;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "xterm-color" "20260531.1854"
  "ANSI, XTERM 256 and Truecolor support."
  '((emacs "25.1"))
  :url "https://github.com/atomontage/xterm-color"
  :commit "0b0d808f8bc5007037341dc5f63149cc32cf2c5b"
  :revdesc "0b0d808f8bc5"
  :keywords '("faces")
  :authors '(("xristos" . "xristos@sdf.org"))
  :maintainers '(("xristos" . "xristos@sdf.org")))
